/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * glx/gltb.c - 17/03/2000
 *
 * Simulation d'un trackball pour la rotation, l'angle de prise de vue
 * et l'loignement.
*/

#include <heliumGL.h>


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Initialise les rotations
*/

void HeTbInitRotations (He_trackball *tb)
{
    tb->beginx = tb->beginy = 0;
    HeTb_trackball(tb->quat, 0.0, 0.0, 0.0, 0.0);
}


/*
 * Initialise l'angle de prise de vue, entre 0.0 et 180.0 degrs
*/

void HeTbInitFovy (He_trackball *tb, GLdouble val, GLdouble min, GLdouble max)
{
    tb->fovy_val = val;
    tb->fovy_min = min;
    tb->fovy_max = max;
}


/*
 * Initialise la distance de l'observateur par rapport  la scne
*/

void HeTbInitZdist (He_trackball *tb, GLdouble val, GLdouble min, GLdouble max)
{
    tb->zdist_val = val;
    tb->zdist_min = min;
    tb->zdist_max = max;
}


/*
 * Mmorise la position de la souris
*/

void HeTbMemoPointer (He_trackball *tb, He_event *hev)
{
    tb->beginx = hev->sx;
    tb->beginy = hev->sy;
}


/*
 * Renvoie l'angle de prise de vue pour gluPerspective()
*/

GLdouble HeTbGetFovy (He_trackball *tb)
{
    return tb->fovy_val;
}


/*
 * Renvoie la distance  l'observateur pour glTranslated()
*/

GLdouble HeTbGetZdist (He_trackball *tb)
{
    return tb->zdist_val;
}


/*
 * Applique les rotations du trackball en multipliant la matrice courante
*/

void HeTbMultRotations (He_trackball *tb)
{
    GLdouble m[4][4];
    
    HeTb_build_rotmatrix (m, tb->quat);
    glMultMatrixd(&m[0][0]);
}


/*
 * Utilise l'vnement pour les rotations du trackball
*/

void HeTbEventRotations (He_trackball *tb, He_node *hn, He_event *hev)
{
    int x = hev->sx, y = hev->sy,
        w = HeGetWidth(hn), h = HeGetHeight(hn);
    GLdouble spin_quat[4];
    
    HeTb_trackball(spin_quat,
              (2.0*tb->beginx - w) / w,
              (h - 2.0*tb->beginy) / h,
              (2.0*x - w) / w,
              (h - 2.0*y) / h );
    HeTb_add_quats(spin_quat, tb->quat, tb->quat);

    HeTbMemoPointer (tb, hev);
}


/*
 * Utilise l'vnement pour l'angle de prise de vue
*/

void HeTbEventFovy (He_trackball *tb, He_node *hn, He_event *hev)
{
    int y = hev->sy, h = HeGetHeight(hn);
 
    tb->fovy_val += ((y - tb->beginy) / h) * 40;
    if (tb->fovy_val < tb->fovy_min) tb->fovy_val = tb->fovy_min;
    if (tb->fovy_val > tb->fovy_max) tb->fovy_val = tb->fovy_max;
    
    HeTbMemoPointer (tb, hev);
}


/*
 * Utilise l'vnement pour la distance  l'observateur
*/

void HeTbEventZdist (He_trackball *tb, He_node *hn, He_event *hev)
{
    int y = hev->sy, h = HeGetHeight(hn);
 
    tb->zdist_val += ((y - tb->beginy) / h) * (tb->zdist_max - tb->zdist_min);
    if (tb->zdist_val < tb->zdist_min) tb->zdist_val = tb->zdist_min;
    if (tb->zdist_val > tb->zdist_max) tb->zdist_val = tb->zdist_max;

    HeTbMemoPointer (tb, hev);
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Trackball code:
 *
 * Implementation of a virtual trackball.
 * Implemented by Gavin Bell, lots of ideas from Thant Tessman and
 *   the August '88 issue of Siggraph's "Computer Graphics," pp. 121-129.
 *
 * Vector manip code:
 *
 * Original code from:
 * David M. Ciemiewicz, Mark Grossman, Henry Moreton, and Paul Haeberli
 *
 * Much mucking with by:
 * Gavin Bell
 *
 * Code adapt  Helium par Edouard Thiel.
*/

/*
 * This size should really be based on the distance from the center of
 * rotation to the point on the object underneath the mouse.  That
 * point would then track the mouse as closely as possible.  This is a
 * simple example, though, so that is left as an Exercise for the
 * Programmer.
*/
#define TRACKBALLSIZE  (0.8)

void HeTb_vzero (GLdouble *v)
{
    v[0] = 0.0;
    v[1] = 0.0;
    v[2] = 0.0;
}

void HeTb_vset (GLdouble *v, GLdouble x, GLdouble y, GLdouble z)
{
    v[0] = x;
    v[1] = y;
    v[2] = z;
}

void HeTb_vsub (GLdouble *src1, GLdouble *src2, GLdouble *dst)
{
    dst[0] = src1[0] - src2[0];
    dst[1] = src1[1] - src2[1];
    dst[2] = src1[2] - src2[2];
}

void HeTb_vcopy (GLdouble *v1, GLdouble *v2)
{
    int i;
    for (i = 0 ; i < 3 ; i++)
        v2[i] = v1[i];
}

void HeTb_vcross (GLdouble *v1, GLdouble *v2, GLdouble *cross)
{
    GLdouble temp[3];

    temp[0] = (v1[1] * v2[2]) - (v1[2] * v2[1]);
    temp[1] = (v1[2] * v2[0]) - (v1[0] * v2[2]);
    temp[2] = (v1[0] * v2[1]) - (v1[1] * v2[0]);
    HeTb_vcopy(temp, cross);
}

GLdouble HeTb_vlength (GLdouble *v)
{
    return sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
}

void HeTb_vscale (GLdouble *v, GLdouble div)
{
    v[0] *= div;
    v[1] *= div;
    v[2] *= div;
}

void HeTb_vnormal (GLdouble *v)
{
    HeTb_vscale(v,1.0/HeTb_vlength(v));
}

GLdouble HeTb_vdot (GLdouble *v1, GLdouble *v2)
{
    return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

void HeTb_vadd (GLdouble *src1, GLdouble *src2, GLdouble *dst)
{
    dst[0] = src1[0] + src2[0];
    dst[1] = src1[1] + src2[1];
    dst[2] = src1[2] + src2[2];
}


/*
 * Ok, simulate a track-ball.  Project the points onto the virtual
 * trackball, then figure out the axis of rotation, which is the cross
 * product of P1 P2 and O P1 (O is the center of the ball, 0,0,0)
 * Note:  This is a deformed trackball-- is a trackball in the center,
 * but is deformed into a hyperbolic sheet of rotation away from the
 * center.  This particular function was chosen after trying out
 * several variations.
 *
 * It is assumed that the arguments to this routine are in the range
 * (-1.0 ... 1.0)
*/
 
void HeTb_trackball (GLdouble q[4], GLdouble p1x, GLdouble p1y, 
                                    GLdouble p2x, GLdouble p2y)
{
    GLdouble a[3]; /* Axis of rotation */
    GLdouble phi;  /* how much to rotate about axis */
    GLdouble p1[3], p2[3], d[3];
    GLdouble t;

    if (p1x == p2x && p1y == p2y) {
        /* Zero rotation */
        HeTb_vzero(q);
        q[3] = 1.0;
        return;
    }

    /*
     * First, figure out z-coordinates for projection of P1 and P2 to
     * deformed sphere
    */
    HeTb_vset(p1,p1x,p1y,HeTb_tb_project_to_sphere(TRACKBALLSIZE,p1x,p1y));
    HeTb_vset(p2,p2x,p2y,HeTb_tb_project_to_sphere(TRACKBALLSIZE,p2x,p2y));

    /*
     *  Now, we want the cross product of P1 and P2
    */
    HeTb_vcross(p2,p1,a);

    /*
     *  Figure out how much to rotate around that axis.
    */
    HeTb_vsub(p1,p2,d);
    t = HeTb_vlength(d) / (2.0*TRACKBALLSIZE);

    /*
     * Avoid problems with out-of-control values...
    */
    if (t > 1.0) t = 1.0;
    if (t < -1.0) t = -1.0;
    phi = 2.0 * asin(t);

    HeTb_axis_to_quat(a,phi,q);
}


/*
 *  Given an axis and angle, compute quaternion.
*/

void HeTb_axis_to_quat (GLdouble a[3], GLdouble phi, GLdouble q[4])
{
    HeTb_vnormal(a);
    HeTb_vcopy(a,q);
    HeTb_vscale(q,sin(phi/2.0));
    q[3] = cos(phi/2.0);
}


/*
 * Project an x,y pair onto a sphere of radius r OR a hyperbolic sheet
 * if we are away from the center of the sphere.
*/

GLdouble HeTb_tb_project_to_sphere (GLdouble r, GLdouble x, GLdouble y)
{
    GLdouble d, t, z;

    d = sqrt(x*x + y*y);
    if (d < r * 0.70710678118654752440) {    /* Inside sphere */
        z = sqrt(r*r - d*d);
    } else {           /* On hyperbola */
        t = r / 1.41421356237309504880;
        z = t*t / d;
    }
    return z;
}


/*
 * Given two rotations, e1 and e2, expressed as quaternion rotations,
 * figure out the equivalent single rotation and stuff it into dest.
 *
 * This routine also normalizes the result every RENORMCOUNT times it is
 * called, to keep error from creeping in.
 *
 * NOTE: This routine is written so that q1 or q2 may be the same
 * as dest (or each other).
*/

#define RENORMCOUNT 97

void HeTb_add_quats (GLdouble q1[4], GLdouble q2[4], GLdouble dest[4])
{
    int count=0;
    GLdouble t1[4], t2[4], t3[4];
    GLdouble tf[4];

    HeTb_vcopy(q1,t1);
    HeTb_vscale(t1,q2[3]);

    HeTb_vcopy(q2,t2);
    HeTb_vscale(t2,q1[3]);

    HeTb_vcross(q2,q1,t3);
    HeTb_vadd(t1,t2,tf);
    HeTb_vadd(t3,tf,tf);
    tf[3] = q1[3] * q2[3] - HeTb_vdot(q1,q2);

    dest[0] = tf[0];
    dest[1] = tf[1];
    dest[2] = tf[2];
    dest[3] = tf[3];

    if (++count > RENORMCOUNT) {
        count = 0;
        HeTb_normalize_quat(dest);
    }
}


/*
 * Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0
 * If they don't add up to 1.0, dividing by their magnitued will
 * renormalize them.
 *
 * Note: See the following for more information on quaternions:
 *
 * - Shoemake, K., Animating rotation with quaternion curves, Computer
 *   Graphics 19, No 3 (Proc. SIGGRAPH'85), 245-254, 1985.
 * - Pletinckx, D., Quaternion calculus as a basic tool in computer
 *   graphics, The Visual Computer 5, 2-13, 1989.
*/

void HeTb_normalize_quat (GLdouble q[4])
{
    int i;
    GLdouble mag;

    mag = (q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
    for (i = 0; i < 4; i++) q[i] /= mag;
}


/*
 * Build a rotation matrix, given a quaternion rotation.
 *
*/

void HeTb_build_rotmatrix (GLdouble m[4][4], GLdouble q[4])
{
    m[0][0] = 1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]);
    m[0][1] = 2.0 * (q[0] * q[1] - q[2] * q[3]);
    m[0][2] = 2.0 * (q[2] * q[0] + q[1] * q[3]);
    m[0][3] = 0.0;

    m[1][0] = 2.0 * (q[0] * q[1] + q[2] * q[3]);
    m[1][1]= 1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]);
    m[1][2] = 2.0 * (q[1] * q[2] - q[0] * q[3]);
    m[1][3] = 0.0;

    m[2][0] = 2.0 * (q[2] * q[0] - q[1] * q[3]);
    m[2][1] = 2.0 * (q[1] * q[2] + q[0] * q[3]);
    m[2][2] = 1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]);
    m[2][3] = 0.0;

    m[3][0] = 0.0;
    m[3][1] = 0.0;
    m[3][2] = 0.0;
    m[3][3] = 1.0;
}

