/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/bubble.c - 29/01/2002
 *
 * Affichage des bulles d'aide.
*/

#include <helium.h>


/* Type et union du package 
*/
#define  OBJ_TYPE	Obj_bubble
#define  OBJ_UNION	obj.bubble

static He_method he_meth_bubble;

#define HE_BUBBLE_XGAP 20
#define HE_BUBBLE_YGAP 8
#define HE_BUBBLE_BORD 3


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 *  Cration d'un bubble.
 *  Renvoie le noeud ou NULL.
*/

He_node *HeCreateBubble ()
{
    He_node *hn;
    OBJ_TYPE *ob;

    HeBubbleAutoInit();

    /* On cree le noeud */
    hn = HeBaseCreate (NULL, &he_meth_bubble);
    if (hn == NULL) return NULL;

    /* On cree l'objet */
    ob = hn->OBJ_UNION = malloc (sizeof (OBJ_TYPE));
    if (ob == NULL) {
	HeError ("HeCreateBubble: malloc error\n");
	return NULL;
    }

    /* Init des champs du noeud */
    HE_BIT_CLR (hn->status, HE_STA_SHOW);
    hn->win_border = 1;
    
    /* init des champs de l'objet */
    ob->label = NULL;
    
    /* Creation du Window avec Xlib */
    hn->win = XCreateSimpleWindow (he_display, 
	he_root_window,
	hn->xb, hn->yb, hn->xm, hn->ym, 
	hn->win_border,
	he_black,
	he_col_bubble);

    /* Proprit du window */
    XSetStandardProperties (he_display, hn->win, "No name", "No name",
	0, he_argv, he_argc, NULL);
        
    HeWMProtocols (hn);
    HeWMClass (hn); 
    HeWMOverrideRedirect (hn);
    HeWMSaveUnder (hn);
    HeEventMask (hn->win);
    HeWinListInsert (hn);

    /* Update diffr */
    HeUpdateMark (hn, HE_UPD_DRAW);

    /* Fin succes */
    return hn;
}


/*
 * Mmorise le texte de la bulle ; il peut contenir des '\n'.
*/

void HeSetBubbleLabel (He_node *hn, const char *label)
{
    OBJ_TYPE *ob;
    
    if (HeNotBubble(hn)) return;
    ob = hn->OBJ_UNION;
    
    free (ob->label);
    ob->label = (label == NULL) ? NULL : strdup(label);

    HeBubbleFit (hn);
}


/*
 * Renvoie le texte de la bulle.
*/

char *HeGetBubbleLabel (He_node *hn)
{
    if (HeNotBubble(hn)) return he_empty_str;
    return hn->OBJ_UNION->label ? hn->OBJ_UNION->label : he_empty_str;
}


/*
 * Affiche le bubble hn aux coordonnes x,y par rapport  ref,
 * ou par rapport  l'cran si ref est NULL.
 * Si flag est TRUE, ignore y et place bubble en dessous (ou au dessus) de hn.
 * Recadre le bubble en x pour qu'il ne dpasse pas l'cran.
*/

void HeShowBubbleXYF (He_node *hn, He_node *ref, int x, int y, int flag)
{
    int ax, ay, xm1, ym1;
    
    if (HeNotBubble(hn)) return;
    
    xm1 = hn->xm + hn->win_border*2;
    ym1 = hn->ym + hn->win_border*2;

    /* Le pre du Window du bubble est le root Window ; il faut donc 
       exprimer les coordonnes x,y par rapport au root Window. */
    HeAbsoluteCoords (ref, x, y, &ax, &ay);

    if (flag) {  /* ignore y, met dessous ou dessus hn */
    
	int ym2 = ref ? ref->ym + ref->win_border*2 : 0;
        ay += HE_BUBBLE_YGAP + ym2 - y;
        if (ay + ym1 > he_root_height) {
            ay -= ym1 + 2*HE_BUBBLE_YGAP + ym2;
            if (ay < 0) ay = 0;
        }
    
    } else {     /* met dessous ou dessus y */
    
        ay += HE_BUBBLE_YGAP;
        if (ay + ym1 > he_root_height) {
            ay -= ym1 + 2*HE_BUBBLE_YGAP;
            if (ay < 0) ay = 0;
        }
    }
    
    /* On recadre en x */
    ax += HE_BUBBLE_XGAP;
    /* ax -= xm1/2 */
    if (ax < 0) ax = 0;
    else if (ax + xm1 > he_root_width) ax = he_root_width - xm1;
    
    /* Affichage en x,y */
    HeSetX (hn, ax); HeSetY (hn, ay); HeSetShow (hn, TRUE);
}


/*-------------------- M E T H O D E S - P R I V E E S ------------------------
 *
 * On ne teste pas si hn == NULL.
*/

/*
 * Initialisation du module ; appel par HeCreateBubble().
*/

void HeBubbleAutoInit ()
{
    static int did_init = FALSE;
    
    if (did_init) return;
    
    HeMethodDefault (&he_meth_bubble, HE_NAT_BUBBLE);
    he_meth_bubble.props   = HE_PM_WIN_OWNER | HE_PM_TOP_WIN;
    he_meth_bubble.destroy = HeBubbleDestroy;
    he_meth_bubble.fit     = HeBubbleFit;
    he_meth_bubble.update  = HeBubbleUpdate;
    he_meth_bubble.event   = HeBubbleEvent;
    he_meth_bubble.draw    = HeBubbleDraw;
    
    did_init = TRUE;
}


/*
 * Destruction d'un bubble ; appele' par HeDestroyRec().
*/

void HeBubbleDestroy (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;

    HeWinListRemove (hn);
    XDestroyWindow (he_display, hn->win);

    free (ob->label);
    
    free (ob);
}


/*
 * Ajuste la taille du bubble en fonction de son contenu.
*/

void HeBubbleFit (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    char *v;
    int c, n, f_a, f_d, w, wmax;
    
    if (HeNotBubble(hn)) return;
    
    v = ob->label;
    if (v == NULL) v = he_empty_str;

    f_a = he_small_font->ascent;
    f_d = he_small_font->descent;
    
    /* Calcul largeur */
    n = 0; wmax = 0;
    while (v[0] != 0)
    {
	c = 0;
	while (v[c] != 0 && v[c] != '\n') c++;
        
        if (c > 0) {
	    w = XTextWidth (he_small_font, v, c);
            if (w > wmax) wmax = w;
            v += c;
        }
        if (v[0] == '\n') { v++; n++; }
    }
    
    HeSetHeight(hn, 2*HE_BUBBLE_BORD + (n+1)*(f_a+f_d));
    HeSetWidth (hn, 2*HE_BUBBLE_BORD + wmax);
    HeUpdateMark (hn, HE_UPD_DRAW);
}


/*
 * M-A-J geometrie et affichage ; appele' par HeUpdateFlush().
*/

void HeBubbleUpdate (He_node *hn, He_upd_info *ui)
{
    HeUpdateWinGeom (hn, ui);
        
    if (HE_BIT_ISSET (ui->mask, HE_UPD_BORDER)) {
	XSetWindowBorderWidth (he_display, hn->win, hn->win_border);
    }

    if (HE_BIT_ISSET (ui->mask, HE_UPD_DRAW)) {
	if (HE_BIT_ISSET (hn->status, HE_STA_SHOW)) {
	    XMapRaised   (he_display, hn->win);
            HeSendExpose (hn->win);
	} else XUnmapWindow (he_display, hn->win);
    }
}


/*
 * Methode pour gerer les evenements X.
 * Renvoie -1 si erreur.
 *
 * Appele' par HeEventDispatch()
*/

int HeBubbleEvent (He_node *hn, He_event *hev)
{
    /* if (hev->xev->type != MotionNotify)
        printf ("bubble %s\n", HeGetXEventName (hev->xev->type)); */

    switch (hev->type) {
    
        case ButtonRelease : HeSetShow (hn, FALSE); break;
        case Expose : HeBubbleDraw (hn); break;

    } /* switch */

    return 0;
}


/*
 * Dessin du bubble.
 *
*/

void HeBubbleDraw (He_node *hn)
{
    OBJ_TYPE *ob = hn->OBJ_UNION;
    char *v;
    int c, n, f_a, f_d;

    v = ob->label;
    if (v == NULL) return;

    HeDrawBg (hn, he_col_bubble);

    XSetForeground (he_display, he_gc_draw, he_black);
    XSetFont (he_display, he_gc_draw, he_small_font->fid);
    f_a = he_small_font->ascent;
    f_d = he_small_font->descent;
    
    n = 0;
    while (v[0] != 0)
    {
	c = 0;
	while (v[c] != 0 && v[c] != '\n') c++;
        
        if (c > 0) {
	    XDrawString (he_display, hn->win, he_gc_draw,
		HE_BUBBLE_BORD, HE_BUBBLE_BORD + f_a + n*(f_a + f_d), v, c );
            v += c;
        }
        if (v[0] == '\n') { v++; n++; }
    }
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 *  Teste nature du noeud ; renvoie TRUE si erreur. 
*/

int HeNotBubble (He_node *hn)
{
    if (hn == NULL || hn->meth != &he_meth_bubble) {
	HeError ("Bubble expected\n");
        return TRUE;
    }
    return FALSE;
}


