/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/color.c - 25/08/1999
 *
 * 
*/

#include <helium.h>

/* 
 * Variables globales et exportees
*/

int he_black, he_white, he_col_bg, he_col_rev_bg, he_col_dark_bd, 
    he_col_light_bd, he_col_ink, he_col_rev_ink, he_col_sel, he_col_toggle,
    he_col_bubble;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Convertions de formats de couleurs
 *
 * Reference: Hearn/Baker, "Computer Graphics", Prentice Hall, Section 14-4.
*/

void HeHsvToRgb (int h, int s, int v, int *r, int *g, int *b)
{
    int i, f, p, q, t;

    s = (s * HE_COLOR_MAXRGB) / HE_COLOR_MAXSV;
    v = (v * HE_COLOR_MAXRGB) / HE_COLOR_MAXSV;
    if (s == 0) { *r = *g = *b = v; return; }
    
    if (h == 360) h = 0;
    i = h / 60;
    f = h % 60;
    p = v * (HE_COLOR_MAXRGB - s) / HE_COLOR_MAXRGB;
    q = v * (HE_COLOR_MAXRGB - s * f / 60) / HE_COLOR_MAXRGB;
    t = v * (HE_COLOR_MAXRGB - s * (60 - f) / 60) / HE_COLOR_MAXRGB;

    switch (i) {
	case 0: *r = v; *g = t; *b = p; break;
	case 1: *r = q; *g = v; *b = p; break;
	case 2: *r = p; *g = v; *b = t; break;
	case 3: *r = p; *g = q; *b = v; break;
	case 4: *r = t; *g = p; *b = v; break;
	case 5: *r = v; *g = p; *b = q; break;
    }
}


void HeRgbToHsv (int r, int g, int b, int *h, int *s, int *v)
{
    int maxv, minv, diffv, hex = 0, rc, gc, bc;
    
    maxv = minv = r;
    if (maxv < g) maxv = g; if (maxv < b) maxv = b; 
    if (minv > g) minv = g; if (minv > b) minv = b;
    diffv = maxv - minv;

    *v = (maxv * HE_COLOR_MAXSV) / HE_COLOR_MAXRGB;
    *s = (maxv) ? ( (diffv * HE_COLOR_MAXRGB / maxv) 
                    * HE_COLOR_MAXSV ) / HE_COLOR_MAXRGB : 0;

    if (*s == 0) { *h = 0; return; }

    rc = (maxv - r) * HE_COLOR_MAXRGB / diffv;
    gc = (maxv - g) * HE_COLOR_MAXRGB / diffv;
    bc = (maxv - b) * HE_COLOR_MAXRGB / diffv;

    if (r == maxv) {
        *h = bc - gc; hex = 0;
    } else if (g == maxv) {
        *h = rc - bc; hex = 2;
    } else if (b == maxv) {
        *h = gc - rc; hex = 4;
    }
    *h = hex * 60 + (*h * 60 / HE_COLOR_MAXRGB);
    if (*h < 0) *h += 360;
}


void HeRgbToXColor (int r, int g, int b, XColor *x)
{
    x->red   = (Ushort) r << 8;
    x->green = (Ushort) g << 8;
    x->blue  = (Ushort) b << 8;
    x->flags = DoRed | DoGreen | DoBlue;
}

void HeHsvToXColor (int h, int s, int v, XColor *x)
{
    int r, g, b;
    HeHsvToRgb (h, s, v, &r, &g, &b);
    HeRgbToXColor (r, g, b, x);
}

void HeXColorToRgb (XColor *x, int *r, int *g, int *b)
{
    *r = x->red   >> 8;
    *g = x->green >> 8;
    *b = x->blue  >> 8;
}

void HeXColorToHsv (XColor *x, int *h, int *s, int *v)
{
    int r, g, b;
    HeXColorToRgb (x, &r, &g, &b);
    HeRgbToHsv (r, g, b, h, s, v);
}


/*
 * Trouve une couleur d'aprs son nom.
 * Renvoie -1 erreur, 0 succes.
*/

int HeParseXColor (const char *name, XColor *x)
{
    int err = XParseColor (he_display, he_default_cmap, name, x);
    return (err == 0) ? -1 : 0;
}

int HeParseRgb (const char *name, int *r, int *g, int *b)
{
    XColor x;
    int err = HeParseXColor (name, &x);
    HeXColorToRgb (&x, r, g, b);
    return err;
}

int HeParseHsv (const char *name, int *h, int *s, int *v)
{
    XColor x;
    int err = HeParseXColor (name, &x);
    HeXColorToHsv (&x, h, s, v);
    return err;
}


/*
 * Alloue une couleur read/only
 *
 * Renvoie l'id de la couleur, sinon defaut
*/

int HeAllocRgb (int r, int g, int b, int defaut)
{
    XColor x;
    HeRgbToXColor (r, g, b, &x);

    if (!XAllocColor (he_display, he_default_cmap, &x)) {
	HeError ("HeAllocRgb(%d,%d,%d): can't alloc color cell\n",
		r, g, b);
	return defaut;
    }
    return x.pixel;
}

int HeAllocHsv (int h, int s, int v, int defaut)
{
    XColor x;
    HeHsvToXColor (h, s, v, &x);

    if (!XAllocColor (he_display, he_default_cmap, &x)) {
	HeError ("HeAllocHsv(%d,%d,%d): can't alloc color cell\n",
		h, s, v);
	return defaut;
    }
    return x.pixel;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialisation des couleurs systeme
 *
*/

void HeColorSysInit ()
{
    int r, g, b, h, s, v;
    char *nom;
    
    /*
     * On peut determiner la visual_class parmi (DirectColor, TrueColor,
     * PseudoColor, StaticColor, GrayScale, StaticGray) en utilisant
     * XMatchVisualInfo(). cf vol.1 p.194
     * RQ certains serveurs sont TrueColor sans etre PseudoColor.
     * Ces informations sont utiles qd on veut savoir si on a le droit
     * d'allouer des cellules read/write. Ici on ne veut que des r/o.
     * 
     * Finalement on ne teste que couleur / noir&blanc ou` depth == 1.
    */

    /* 
     * Dans tous les cas 
    */
    he_black = BlackPixel (he_display, he_screen);
    he_white = WhitePixel (he_display, he_screen);

    /* 
     * Mode noir et blanc
    */
    if (he_default_depth == 1) {
	he_col_bg       = he_white;
	he_col_rev_bg   = he_black;
	he_col_dark_bd  = he_black;
	he_col_light_bd = he_black;
	he_col_ink      = he_black;
	he_col_rev_ink  = he_white;
	he_col_sel      = he_black;
	he_col_toggle   = he_black;
        he_col_bubble	= he_white;
	return;
    }

    /* 
     * Mode couleur
    */
    
    /* Couleur de reference du window */
    
    nom = HeGetOption ("-color", "#cccccc");
    if (HeParseRgb (nom, &r, &g, &b) != 0) { 
	HeError ("HeColorSysInit: can't get window color \"%s\"\n", nom);
	r = 204; g = 204; b = 204;
    }
    he_col_bg = HeAllocRgb (r, g, b, he_white);
    
    /* Calcul des couleurs 3D a partir de la couleur de reference */
    
    HeRgbToHsv (r, g, b, &h, &s, &v);
    he_col_rev_bg  = HeAllocHsv (h, s, (v*9)/10, he_black);  /* 90% */
    he_col_dark_bd = HeAllocHsv (h, s, (v*5)/10, he_black);  /* 50% */

    v = (v*12)/10; /* 120% */
    if (v > HE_COLOR_MAXSV) { v = HE_COLOR_MAXSV; s /= 2; }
    if (v < (HE_COLOR_MAXSV*4)/10) v = (HE_COLOR_MAXSV*4)/10;
    he_col_light_bd = HeAllocHsv (h, s, v, he_black);
    
    /* Couleur de la selection */
    
    nom = HeGetResource ("selectColor", "#F0F000");
    if (HeParseRgb (nom, &r, &g, &b) != 0) { 
	HeError ("HeColorSysInit: can't get selection color \"%s\"\n", nom);
	r = 240; g = 240; b = 0;
    }
    he_col_sel = HeAllocRgb (r, g, b, he_black);

    /* Couleur des boutons toggle */
    
    nom = HeGetResource ("toggleColor", "green");
    if (HeParseRgb (nom, &r, &g, &b) != 0) { 
	HeError ("HeColorSysInit: can't get toggle button color \"%s\"\n", nom);
	r = 0; g = 255; b = 0;
    }
    he_col_toggle = HeAllocRgb (r, g, b, he_black);

    /* Couleur de l'encre */
    
    he_col_ink     = he_black;
    he_col_rev_ink = (he_col_rev_bg == he_black) ? he_white : he_black;
    
    /* Couleur de fond pour les bulles */
    he_col_bubble = HeAllocRgb (248, 248, 208, he_white);
    
    return ;   
}

