/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/comm.c - 27/10/1997
 *
 * Appels systeme "portables"
*/

#include <helium.h>


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Creation d'un socket dans le domaine AF_INET (ip).
 *
 * IN	type = SOCK_DGRAM (udp), SOCK_STREAM (tcp)
 *	port = numero de port entier, ou 0 pour allocation auto.
 *
 * OUT	port = numero de port effectif
 *	addr = adresse effective, contenant le port.
 *
 * RET	descripteur socket, ou -1 erreur.
 *
 * Apres l'appel de la fonction, l'adresse est localhost ;
 * pour adresser une machine distante, faire :
 *
 *	struct hostent *hp;
 *
 *	if ((hp = gethostbyname (machine)) == NULL)
 *	     fprintf (stderr, "`%s' : Host unknown\n", machine);
 *	else memcpy (&addr.sin_addr, hp->h_addr, hp->h_length);
*/

int HeCreateSocket (int type, int *port, struct sockaddr_in *addr)
{
    int fd;
    socklen_t longueur = sizeof (struct sockaddr_in);

    if (port == NULL || addr == NULL) {
	HeError ("HeCreateSocket: can't create socket: null pointers\n");
	return -1;
    }

    /* Creation de la socket */
    if ((fd = socket (AF_INET, type, 0)) < 0) {
	HeError ("HeCreateSocket: can't create socket: %s\n", strerror(errno));
	return -1;
    }

    /* Preparation de l'adresse d'attachement */
    memset ((char *) addr, 0, longueur);
    addr->sin_family = AF_INET;
    addr->sin_port = htons (*port);  /* forme Network */
    addr->sin_addr.s_addr = htonl(INADDR_ANY);	/* localhost */

    /* Demande d'attachement de la socket */
    if (bind (fd, (struct sockaddr *) addr, longueur) < 0) {
	HeError ("HeCreateSocket: can't bind socket: %s\n", strerror(errno));
	close (fd);
	return -1;
    }

    /* Recuperation de l'adresse effective d'attachement */
    if (getsockname (fd, (struct sockaddr *) addr, &longueur) < 0) {
	HeError ("HeCreateSocket: can't get socket name: %s\n", 
        	strerror(errno));
	close (fd);
	return -1;
    }
    *port = ntohs (addr->sin_port);  /* forme Host */

    return fd;
}


/*
 * Rend un fd asynchrone, avec envoi du signal SIGIO.
 *
 * Retour 0 succes, -1 erreur.
*/

int HeSetAsync (int fd)
{

#ifndef FASYNC
    int on = 1;
#endif

    /* attribution du fd au processus */
    if (fcntl (fd, F_SETOWN, getpid()) == -1) {
	HeError ("HeSetAsync: can't set owner of fd: %s\n", strerror(errno));
	return -1;
    }

    /* activation du mecanisme d'envoi de SIGIO */
#ifdef FASYNC
    if (fcntl (fd, F_SETFL, fcntl(fd, F_GETFL) | FASYNC) == -1) {
#else
    if (ioctl (fd, FIOASYNC, &on) == -1) {
#endif
	HeError ("HeSetAsync: can't activate SIGIO: %s\n", strerror(errno));
	return -1;
    }
    return 0;
}


/*
 * Rend un fd non bloquant ou bloquant.
 *
 * Retour 0 succes, -1 erreur.
*/

int HeSetNonblock (int fd, int nonblock)
{
    int flag = (nonblock)
	? fcntl(fd, F_GETFL) | O_NONBLOCK
	: fcntl(fd, F_GETFL) & ~O_NONBLOCK;

    /* fd bloquant ou non */
    if (fcntl (fd, F_SETFL, flag) == -1) {
	HeError ("HeSetNonblock: can't change blocking mode of fd: %s\n", 
        	strerror(errno));
	return -1;
    }

    return 0;
}


