/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/draw.c - 25/08/1999
 *
 *
 * RAPPEL IMPORTANT :
 *   un rectangle (xb, yb, xm, ym) a pour coin droit (xb+xm-1, yb+ym-1). 
 *
 *   XFillRectangle (..., xb, yb, xm, ym);
 *   XDrawRectangle (..., xb, yb, xm-1, ym-1);
 *   HeDrawRect..   (..., xb, yb, xm-1, ym-1, ...);
*/

#include <helium.h>


/* 
 * Variables globales et exportees
*/

char he_empty_str[] = "";
XFontStruct *he_normal_font, *he_bold_font, *he_small_font;
int he_font_width, he_font_height;
GC he_gc, he_gc_draw, he_gc_xor;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Dessine un rectangle avec le look 3D
 *
 * Exemple d'appel : HeDrawRect3D (win, 
 *			hn->xb, hn->yb, hn->xm-1, hn->ym-1,
 *			he_col_dark_bd, he_col_light_bd);
*/

void HeDrawRect3D (Window win, int xb, int yb, int xm, int ym, int c1, int c2)
{
    int x2 = xb+xm, y2 = yb+ym;

    XSetForeground (he_display, he_gc_draw, c1);
    XDrawLine (he_display, win, he_gc_draw, xb, yb, x2, yb);
    XDrawLine (he_display, win, he_gc_draw, xb, yb, xb, y2);

    XSetForeground (he_display, he_gc_draw, c2);
    XDrawLine (he_display, win, he_gc_draw, xb+1, y2, x2, y2);
    XDrawLine (he_display, win, he_gc_draw, x2, yb+1, x2, y2);
}


/*
 * Dessine un rectangle avec le look 3D et les coins e'vide's
 *
 * Exemple d'appel : HeDrawRect3DC (win, 
 *			hn->xb, hn->yb, hn->xm-1, hn->ym-1,
 *			he_col_dark_bd, he_col_light_bd);
*/

void HeDrawRect3DC (Window win, int xb, int yb, int xm, int ym, int c1, int c2)
{
    int x2 = xb+xm, y2 = yb+ym;

    XSetForeground (he_display, he_gc_draw, c1);
    XDrawLine (he_display, win, he_gc_draw, xb+1, yb, x2-1, yb);
    XDrawLine (he_display, win, he_gc_draw, xb, yb+1, xb, y2-1);

    XSetForeground (he_display, he_gc_draw, c2);
    XDrawLine (he_display, win, he_gc_draw, xb+1, y2, x2-1, y2);
    XDrawLine (he_display, win, he_gc_draw, x2, yb+1, x2, y2-1);
}


/*
 * Affiche un string ; x,y = coin en haut a gauche
*/

void HeDrawString (Window win, GC gc, XFontStruct *font, 
	int x, int y, const char *ligne)
{
    if (!ligne) return;

    XSetFont (he_display, gc, font->fid);

    XDrawString (he_display, win, gc,
	x, y + font->ascent, 
	ligne, HeStrLen(ligne) );
}


/*
 * Affiche le sous-string [pos .. pos+len[ ; x,y = coin en haut a gauche
*/

void HeDrawSubString (Window win, GC gc, XFontStruct *font, 
	int x, int y, const char *ligne, int pos, int len)
{
    int l;
    
    if (!ligne) return;
    l = HeStrLen(ligne);
    if (pos < 0 || pos >= l) return;
    ligne += pos; l -= pos;
    if (l > len) l = len;

    XSetFont (he_display, gc, font->fid);

    XDrawString (he_display, win, gc,
	x, y + font->ascent, 
	ligne, l );
}


/*
 * Affiche un string centre' dans le rectangle (xb, yb, xm, ym)
*/

void HeDrawStringCenterRect (Window win, GC gc, XFontStruct *font, 
	int xb, int yb, int xm, int ym, const char *ligne)
{
    int l;

    if (!ligne) return;
    l = HeStrLen(ligne);

    XSetFont (he_display, gc, font->fid);

    XDrawString (he_display, win, gc,
	xb + (xm - XTextWidth (font, ligne, l) )/2, 
	yb + (ym + font->ascent - font->descent)/2, 
	ligne, l );
}


/*
 * Affiche un string justifi selon pos.
 *
 * L'equivalent de HeDrawString (.., x, y, ..)
 *          est HeDrawStringPos (.., x, y, HE_BASE_LEFT, ..)
*/

void HeDrawStringPos (Window win, GC gc, XFontStruct *font, 
	int x, int y, int pos, const char *ligne)
{
    int dx, dy, l, w;
    
    if (!ligne) return;
    l = HeStrLen (ligne);
    w = XTextWidth (font, ligne, l);

    XSetFont (he_display, gc, font->fid);
    
    switch (pos) {
      case HE_TOP_LEFT :      dx =  0;   dy = font->ascent;   break;
      case HE_TOP_MIDDLE :    dx = -w/2; dy = font->ascent;   break;
      case HE_TOP_RIGHT :     dx = -w;   dy = font->ascent;   break;
      case HE_BASE_LEFT :     dx =  0;   dy = 0;              break;
      case HE_BASE_MIDDLE :   dx = -w/2; dy = 0;              break;
      case HE_BASE_RIGHT :    dx = -w;   dy = 0;              break;
      case HE_BOTTOM_LEFT :   dx =  0;   dy = -font->descent; break;
      case HE_BOTTOM_MIDDLE : dx = -w/2; dy = -font->descent; break;
      case HE_BOTTOM_RIGHT :  dx = -w;   dy = -font->descent; break;
      default :               dx = 0;    dy = 0;
    }

    XDrawString (he_display, win, gc, x + dx, y + dy, ligne, l );
}


/*
 * Longueur d'un string eventuellement NULL.
*/

int HeStrLen (const char *s)
{
    if (s == NULL) return 0;
    return strlen (s);
}

/*
 * Dimensions a l'ecran d'un string dans une fonte
*/

void HeDimString (XFontStruct *font, const char *ligne, int *x, int *y)
{
    if (!ligne) ligne = he_empty_str;
    *x = XTextWidth (font, ligne, HeStrLen(ligne));
    *y = font->ascent + font->descent;
}


/*
 * Dessin du fond d'un item avec couleur coul
*/

void HeDrawBg (He_node *hn, int coul)
{
    XSetForeground (he_display, he_gc_draw, coul);

    if (HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER))
	/* C'est un widget, on efface tout le window */
	XFillRectangle (he_display, hn->win, he_gc_draw,
		0, 0, hn->xm, hn->ym);
    else
	/* C'est un gadget */
	XFillRectangle (he_display, hn->win, he_gc_draw,
		hn->xb, hn->yb, hn->xm, hn->ym);
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialisation des fontes
*/

int HeInitFonts ()
{
    int i, d;
    char *a_normal, *a_bold, *a_small, *a_size, 
         n_normal[255], n_bold[255], n_small[255];
         
    char *font_normal[] = {
	n_normal,
	"6x13",
        ""	/* marqueur fin liste */
     };
    char *font_bold[] = {
	n_bold,
 	"6x13bold",
        ""	/* marqueur fin liste */
    };
    char *font_small[] = {
	n_small,
 	"6x13",
        ""	/* marqueur fin liste */
    };

    font_normal[0] = n_normal;
    font_bold[0] = n_bold;
    font_small[0] = n_small;
    
    /* 
     * Traitement des options
    */
    a_size = HeGetOption ("-fs", "14");
    d = atoi(a_size);  /* securite overflow pour sprintf */
    sprintf (n_normal, "-adobe-courier-medium-r-normal--%d-*-*-*-*-*-*-*", d);
    sprintf (n_bold,   "-adobe-courier-bold-r-normal--%d-*-*-*-*-*-*-*", d);
    sprintf (n_small,  "-adobe-helvetica-medium-r-normal--%d-*-*-*-*-*-*-*", d-2);
    
    a_normal = HeGetOption ("-fn", NULL);
    if (a_normal) {
        strncpy (n_normal, a_normal, 254); n_normal[254] = 0;
        font_normal[1] = "";  /* marqueur fin liste */
    }
    
    a_bold = HeGetOption ("-fb", NULL);
    if (a_bold) {
        strncpy (n_bold, a_bold, 254); n_bold[254] = 0;
        font_bold[1] = "";  /* marqueur fin liste */
    }
    
    a_small = HeGetOption ("-fl", NULL);
    if (a_small) {
        strncpy (n_small, a_small, 254); n_small[254] = 0;
        font_small[1] = "";  /* marqueur fin liste */
    }
    
    /* 
     * Parcours liste normale
    */
    for (i = 0; font_normal[i][0] ; i++) {
        he_normal_font = XLoadQueryFont (he_display, font_normal[i]);
        if (he_normal_font) break;
    }
    if (he_normal_font == NULL) {
        HeError ("HeInitFonts: can't load normal font; use option -fn\n");
        return -1;
    }
    
    /* 
     * Parcours liste bold
    */
    for (i = 0; font_bold[i][0] ; i++) {
        he_bold_font = XLoadQueryFont (he_display, font_bold[i]);
        if (he_bold_font) break;
    }
    if (he_bold_font == NULL) {
        HeError ("HeInitFonts: can't load bold font; use option -fb\n");
        return -1;
    }
    
    /* 
     * Parcours liste small
    */
    for (i = 0; font_small[i][0] ; i++) {
        he_small_font = XLoadQueryFont (he_display, font_small[i]);
        if (he_small_font) break;
    }
    if (he_small_font == NULL) {
        HeError ("HeInitFonts: can't load small font; use option -fl\n");
        return -1;
    }
    
    /* Taille d'un caractere */
    HeDimString (he_normal_font, "M", &he_font_width, &he_font_height);
 
    return 0;
}


/*
 * Initialisation des GC
*/

void HeInitGCs ()
{
    XGCValues values;

    /* GC public pour API */
    he_gc = DefaultGC (he_display, he_screen);

    /* GC prive' pour dessin normal */
    he_gc_draw = XCreateGC (he_display, he_root_window, 0L, &values);
    XCopyGC (he_display, he_gc, ~0L, he_gc_draw);
    XSetFunction (he_display, he_gc_draw, GXcopy);

    /* GC prive' pour dessin en xor */
    he_gc_xor = XCreateGC (he_display, he_root_window, 0L, &values);
    XCopyGC (he_display, he_gc, ~0L, he_gc_xor);
    XSetFunction (he_display, he_gc_xor, GXxor);
    XSetForeground (he_display, he_gc_xor, 255);
}


