/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/focus.c - 24/01/2002
 *
 * Gestion du focus clavier.
 *
 * Le propritaire du focus est le widget qui reoit les event du clavier.
 * La racine de focus d'un widget est, parmi les parents du widget, celui qui
 * mmorise quel est le propritaire actuel du focus dans le groupe.
 * Le groupe de focus d'un widget est l'ensemble des descendants de la racine
 * de focus du widget.
 *
 * Le propritaire est marqu par le flag HE_STA_FOCUS_OWN.
 * Lorsque la racine reoit le focus du WM, la racine et le propritaire
 * sont marqus par le flag HE_STA_FOCUS_IN.
 * La racine est le parent qui a la proprit HE_PM_STORE_FOCUS.
*/

#include <helium.h>


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Recherche du noeud qui a le focus dans le groupe de focus de hn.
 * Renvoie NULL si hn est NULL.
*/

He_node *HeGetFocus (He_node *hn)
{
    He_node *root;
    He_focus_info *focus_info;
    
    root = HeFocusFindRoot (hn);
    if (root == NULL) return NULL;
    
    focus_info = (root->meth->focus)(root);
    if (focus_info == NULL) return NULL;
    
    return HeGetNodeFromId(focus_info->id);
}


/*
 * Enleve le focus a l'actuel propritaire dans le groupe de hn ;
 * si flag est TRUE, donne le focus a hn.
 * Ne fait rien si hn est NULL.
*/

void HeSetFocus (He_node *hn, int flag)
{
    He_node *root;
    He_focus_info *focus_info;
    
    root = HeFocusFindRoot (hn);
    if (root == NULL) return;
    
    focus_info = (root->meth->focus)(root);
    if (focus_info == NULL) return;
    
    HeSetFocusP (root, focus_info, flag ? hn : NULL);
}


/*
 * Passe le focus au prochain item du groupe de hn.
 * Generalement appele quand [TAB].
 * Ne fait rien si hn est NULL.
*/

He_node *HeNextFocus (He_node *hn)
{
    return HeFocusSearch (hn, 1);
}


/*
 * Passe le focus au precedent item du groupe de hn.
 * Generalement appele quand [Shift TAB].
 * Ne fait rien si hn est NULL.
*/

He_node *HePredFocus (He_node *hn)
{
    return HeFocusSearch (hn, -1);
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialise les champs de focus_info.
*/

void HeFocusInitInfo (He_focus_info *focus_info)
{
    if (focus_info == NULL) return;
    focus_info->id = 0;
    focus_info->borrow = 0;
}


/*
 * Met ou enleve le flag HE_STA_FOCUS_IN au propritaire du focus et  la
 * racine du groupe de focus de hn. Ne fait rien si hn est NULL.
 *
 * Appel par les mthodes He*Event.
*/

void HeFocusReceiveEvent (He_node *hn, int flag)
{
    He_node *root, *focus;
    He_focus_info *focus_info;
    
    /* printf ("HeFocusReceiveEvent %d %s\n",
        hn ? hn->id : 0, flag ? "TRUE" : "FALSE"); */
    
    root = HeFocusFindRoot (hn);
    if (root == NULL) return;
    
    if (flag)
         HE_BIT_SET (root->status, HE_STA_FOCUS_IN);
    else HE_BIT_CLR (root->status, HE_STA_FOCUS_IN);
    
    focus_info = (root->meth->focus)(root);
    if (focus_info == NULL) return;
    
    focus = HeGetNodeFromId(focus_info->id);
    if (focus == NULL) return;
    
    if (flag == (HE_BIT_ISSET (focus->status, HE_STA_FOCUS_IN))) return;
    
    if (flag)
         HE_BIT_SET (focus->status, HE_STA_FOCUS_IN);
    else HE_BIT_CLR (focus->status, HE_STA_FOCUS_IN);
    
    HeUpdateMark (focus, HE_UPD_FOCUS);
    
    /* printf ("HeFocusReceiveEvent done, id = %d bo = %d\n", 
        focus_info->id, focus_info->borrow); */
}


/*
 * Emprunte ou restitue le focus au propritaire.
 * 
 * Appel par les mthodes He*Event.
*/

void HeFocusBorrow (He_node *hn, int flag)
{
    He_node *root;
    He_focus_info *focus_info;
    
    /* printf ("HeFocusBorrow %d %s\n",
	hn->id, flag ? "TRUE":"FALSE"); */
    
    root = HeFocusFindRoot (hn);
    if (root == NULL) return;
    
    focus_info = (root->meth->focus)(root);
    if (focus_info == NULL) return;
    
    /* Renonce si 2 appels de suite avec le mme flag */
    if (flag == (focus_info->id == hn->id)) return;
    
    if (flag) {  /* Emprunte le focus : id -> borrow */
	
        He_node *actu = HeGetNodeFromId(focus_info->id);
        if (actu) {
            HE_BIT_CLR (actu->status, HE_STA_FOCUS_IN);
	    HeUpdateMark (actu, HE_UPD_FOCUS);
        }
        HE_BIT_SET (hn->status, HE_STA_FOCUS_OWN);
        if (HE_BIT_ISSET (root->status, HE_STA_FOCUS_IN))
            HE_BIT_SET (hn->status, HE_STA_FOCUS_IN);
        HeUpdateMark (hn, HE_UPD_FOCUS);
        
        focus_info->borrow = focus_info->id;
        focus_info->id = hn->id;
        
    } else {  /* Restitue le focus : borrow -> id */
	
        He_node *actu = HeGetNodeFromId(focus_info->id);
	He_node *borr = HeGetNodeFromId(focus_info->borrow);
        if (actu) {
            HE_BIT_CLR (actu->status, HE_STA_FOCUS_OWN);
            HE_BIT_CLR (actu->status, HE_STA_FOCUS_IN);
	    HeUpdateMark (actu, HE_UPD_FOCUS);
        }
        if (borr && HE_BIT_ISSET (root->status, HE_STA_FOCUS_IN)) {
            HE_BIT_SET (borr->status, HE_STA_FOCUS_IN);
            HeUpdateMark (borr, HE_UPD_FOCUS);
        }
        focus_info->id = focus_info->borrow;
        focus_info->borrow = 0;
    }
}


/*
 * Recherche de la racine de focus de hn, c'est--dire du parent de hn qui 
 * stocke le propritaire du focus.
 * Renvoie NULL si hn est NULL ou si la racine n'a pas t trouve.
*/

He_node *HeFocusFindRoot (He_node *hn)
{
    He_node *root;
    
    for (root = hn; root != NULL; root = root->owner)
	if (HE_BIT_ISSET (root->meth->props, HE_PM_STORE_FOCUS)) break;
        
    return root;
}


/*
 * Enlve le focus  focus_info->id, et le donne  nouv si nouv != NULL.
 * On suppose que root et focus_info sont non nuls.
*/

void HeSetFocusP (He_node *root, He_focus_info *focus_info, He_node *nouv)
{
    He_node *old = HeGetNodeFromId(focus_info->id);
        
    /* Perte de l'ancien focus */
    if (old != NULL && old != nouv) {
	HE_BIT_CLR (old->status, HE_STA_FOCUS_OWN);
        HE_BIT_CLR (old->status, HE_STA_FOCUS_IN);
        HeUpdateMark (old, HE_UPD_FOCUS);
    }
    
    /* RAZ de borrow (par scurit) */
    if (focus_info->borrow != 0) {
        He_node *bor = HeGetNodeFromId(focus_info->borrow);
        if (bor != NULL) {
	    HE_BIT_CLR (bor->status, HE_STA_FOCUS_OWN);
            HE_BIT_CLR (bor->status, HE_STA_FOCUS_IN);
            HeUpdateMark (bor, HE_UPD_FOCUS);
        }
        focus_info->borrow = 0;
    }
        
    /* Le nouveau focus (il peut etre NULL) */
    if (nouv != old) {
        if (nouv == NULL) { focus_info->id = 0; return; }
        focus_info->id = nouv->id;
        HE_BIT_SET (nouv->status, HE_STA_FOCUS_OWN);
        if (HE_BIT_ISSET (root->status, HE_STA_FOCUS_IN))
            HE_BIT_SET (nouv->status, HE_STA_FOCUS_IN);
        HeUpdateMark (nouv, HE_UPD_FOCUS);
    }
}


/*
 * Recherche recursive du focus suivant (dir = +1) ou precedent (dir = -1).
 * Appele' par HeNextFocus() et HePredFocus().
 *
 * La recherche s'effectue sur un critre geographique, independant de
 * l'ordre de chainage ; on travaille en coordonnees absolues par rapport
 *  la racine de focus (le parent qui stocke le focus).
*/

static He_node *he_candi, *he_focus;
static int he_candi_x, he_candi_y, he_focus_x, he_focus_y;

He_node *HeFocusSearch (He_node *hn, int dir)
{
    He_node *root;
    He_focus_info *focus_info;
    
    root = HeFocusFindRoot (hn);
    if (root == NULL) return NULL;  /* protege aussi du cas hn == NULL */
    
    focus_info = (root->meth->focus)(root);
    if (focus_info == NULL) return NULL;

    /* Init : focus actuel, candidat */
    he_focus = HeGetNodeFromId(focus_info->id);
    he_candi = NULL;
    
    /* Coordonnees absolues du focus / racine du groupe */
    /* RQ on ne tient pas compte du win_border ; il le faudrait. */
    if (he_focus != NULL) {
	He_node *tmp;
        
	he_focus_x = he_focus->xb; he_focus_y = he_focus->yb;
    
	for (tmp = he_focus->owner; tmp != NULL; tmp = tmp->owner)
	    if (HE_BIT_ISSET (tmp->meth->props, HE_PM_STORE_FOCUS)) break;
            else if (HE_BIT_ISSET (tmp->meth->props, HE_PM_WIN_OWNER))
		{ he_focus_x += tmp->xb; he_focus_y += tmp->yb; }
    }
    
    /* Recherche recursive */
    HeFocusSearchRec (root->sub_node, 0, 0, dir);
    
    /* Change le focus */
    HeSetFocusP (root, focus_info, he_candi);
    
    return he_candi;
}

void HeFocusSearchRec (He_node *hn, int xf, int yf, int dir)
{
    int x, y;
    
    for ( ; hn != NULL ; hn = hn->next_node) {
 
	/* Coord absolues hn / racine du groupe */
        x = hn->xb + xf; y = hn->yb + yf;
 
        /* Test si hn est un meilleur candidat au focus */
        if ( ( HE_BIT_ISSET (hn->meth->props, HE_PM_ACCEPT_FOCUS) &&
               HE_BIT_ISSET (hn->status, HE_STA_SHOW) &&
               HE_BIT_ISSET (hn->status, HE_STA_ACTIVE) )
          && ( he_focus == NULL ||
               (dir > 0 && ( y > he_focus_y ||
                            (y == he_focus_y && x > he_focus_x) )) ||
               (dir < 0 && ( y < he_focus_y ||
                            (y == he_focus_y && x < he_focus_x) )) )
          && ( he_candi == NULL ||
               (dir > 0 && ( y < he_candi_y ||
                            (y == he_candi_y && x < he_candi_x) )) ||
               (dir < 0 && ( y > he_candi_y ||
                            (y == he_candi_y && x > he_candi_x) )) ) )
        { he_candi = hn; he_candi_x = x; he_candi_y = y; }

        /* Appel recursif sur le fils ; choix de l'origine relative */
        if ( HE_BIT_ISSET (hn->status, HE_STA_SHOW) &&
             HE_BIT_ISSET (hn->status, HE_STA_ACTIVE) ) {
            if (HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER))
                 HeFocusSearchRec (hn->sub_node, x, y, dir);
            else HeFocusSearchRec (hn->sub_node, xf, yf, dir);
        }
    }
}


/*
 * Envoie l'evenement au focus.
 * S'il n'y a pas de focus, arrete.
 * Si l'evt est [TAB] ou [Shift TAB],
 *     - on passe le focus 
 *     - on transmet cet evt si HE_PM_SEND_TAB.
 *
 * On suppose que hn et hev sont non nuls.
*/

void HeSendEventToFocus (He_node *hn, He_event *hev)
{
    He_node *ga = NULL;
    enum { A_GET, A_NEXT, A_PRED } action = A_GET;
    
    switch (hev->type) {
	case KeyPress :	
            if (hev->sb == 0)
	        switch (hev->sym) {
                    case XK_Tab :
                    case XK_KP_Tab :
                        action = A_NEXT;
                        break;
 
                    case XK_ISO_Left_Tab :
                        action = A_PRED;
                        break;
                }
            break;
    }
    
    switch (action) {
        case A_GET  : ga = HeGetFocus  (hn); break;
        case A_NEXT : ga = HeNextFocus (hn); break;
        case A_PRED : ga = HePredFocus (hn); break;
    }

    if (ga == NULL) return;
    
    if ( (action == A_NEXT || action == A_PRED) &&
         !HE_BIT_ISSET (ga->meth->props, HE_PM_SEND_TAB) ) return;
    
    /* hn est le widget qui a reu l'vnement clavier,
     * ga est le widget  qui on va passer l'vnement.
     * Pour viter une boucle rcursive (He*Event appelle HeSendEventToFocus
     * qui rappelle He*Event sur le mme vnement, etc), on n'appelle 
     * ga->meth->event que si ga != hn. 
    */
    if (ga != hn)
        ga->meth->event (ga, hev);
    else
        HeError ("HeSendEventToFocus: %s dropped\n",
            HeGetXEventName (hev->xev->type));       
}


