/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/io.c - 31/07/2000
 *
 * Ordonnancement des E/S, des evenements X11, des timeout et des signaux 
 * avec la primitive systme select().
 * C'est la partie la plus "profonde" de la gestion des vnements dans
 * le toolkit.
*/

#include <helium.h>


static He_vlist he_io_list;

typedef struct {
    int fd;
    He_io_proc proc;
    void *data;
    Ulong cond, rcond;  /* Masques de conditions autorises, produites */
} He_io_struct;

static int he_ioloop_flag = TRUE;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Ajoute une callback sur un fd.
 * Un nouvel appel ecrase le precedent. 
 * Renvoie 0 succs, -1 erreur.
 *
 * cond est une combinaison bit  bit (avec l'oprateur |) de
 *    HE_IO_READ    : le fd est prt en lecture
 *    HE_IO_WRITE   : le fd est prt en criture
 *    HE_IO_EXCEPT  : le fd reoit une donne urgente
 * c'est la condition pour laquelle la callback est appele.
 *
 * Le prototype de la callback proc est
 *    void proc (int fd, Ulong rcond, void *data);
 * proc est appele chaque fois qu'une des conditions est remplie ; proc reoit
 * fd, rcond et data ; rcond contient les conditions qui sont satisfaites.
*/

int HeAddIO (int fd, Ulong cond, He_io_proc proc, void *data)
{
    int i;
    He_io_struct *elt;
    
    if (fd < 0) {
        HeError ("HeAddIO: fd < 0\n"); 
        return -1;
    }
    if (proc == NULL) {
        HeError ("HeAddIO(%d): proc is NULL\n", fd); 
        return -1;
    }
    
    i = HeIOFind (fd);
    
    /* Insertion ventuelle d'un nouveau fd */
    if (i < 0) {
        He_io_struct tmp;

        tmp.fd = fd;
        i = HeVListInsert (&he_io_list, &tmp);
        if (i < 0) {
            HeError ("HeAddIO(%d): HeVListInsert failed\n", fd); 
            return -1;
        }
    }
    
    elt = &((He_io_struct*)he_io_list.v)[i];
    elt->proc = proc;
    elt->data = data;
    elt->cond = cond;
    elt->rcond = HE_IO_NULL;
    
    return 0;
}


/*
 * Supprime une callback sur un fd.
 * renvoie 0 succs, -1 erreur.
*/

int HeRemoveIO (int fd)
{
    int i;
    
    i = HeIOFind (fd);
    if (i < 0) {
        HeError ("HeRemoveIO(%d): fd not in list\n", fd); 
        return -1;
    }
    
    return HeVListRemove (&he_io_list, i);
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialise le module.
*/

void HeIOInit ()
{
    /* Liste de He_io_struct trie dans l'ordre croissant des fd */
    HeVListInit (&he_io_list, TRUE, 8, sizeof(He_io_struct), HeIOCompar);
}


/*
 * Comparaison des fd pour la vlist
*/

int HeIOCompar (void *v1, void *v2)
{
    return ((He_io_struct*)v1)->fd == ((He_io_struct*)v2)->fd ?  0 :
           ((He_io_struct*)v1)->fd <  ((He_io_struct*)v2)->fd ? -1 : 1;
}


/*
 * Recherche de l'index d'un fd dans la vlist
 * Renvoie >= 0 index, -1 erreur.
*/

int HeIOFind (int fd)
{
    He_io_struct elt;
    int index, res;

    elt.fd = fd;
    res = HeVListSearch (&he_io_list, &elt, &index);
    return (res == 0) ? index : -1;
}


/*
 * Scrutation des E/S avec un select().
 *
 * Attente bloquante avec le timeout tp (infini si tp est NULL).
 *
 * renvoie >0 : nb de descripteurs modifis, 
 *          0 : le timeout a expir,
 *         -1 : appel interrompu par un signal,
 *         -2 : erreur dans select().
*/

int HeIOPoll (struct timeval *tp)
{
    fd_set set_read, set_write, set_except;
    int i, maxfd = 0, res;
    He_io_struct *v = he_io_list.v;  /* non modifi dans la fct */
    
    FD_ZERO (&set_read);
    FD_ZERO (&set_write);
    FD_ZERO (&set_except);
    
    for (i = 0; i < he_io_list.nb; i++) {
        if (HE_BIT_ISSET (v[i].cond, HE_IO_READ))   FD_SET (v[i].fd, &set_read);
        if (HE_BIT_ISSET (v[i].cond, HE_IO_WRITE))  FD_SET (v[i].fd, &set_write);
        if (HE_BIT_ISSET (v[i].cond, HE_IO_EXCEPT)) FD_SET (v[i].fd, &set_except);
	if (v[i].fd > maxfd) maxfd = v[i].fd;
        v[i].rcond = HE_IO_NULL;
    }
    
    /* Appel bloquant avec un timeout tp (infini si tp est NULL) */
    res = select (maxfd+1, &set_read, &set_write, &set_except, tp);
    
    if (res > 0)
    for (i = 0; i < he_io_list.nb; i++) {
	if (FD_ISSET (v[i].fd, &set_read  )) HE_BIT_SET (v[i].rcond, HE_IO_READ);
	if (FD_ISSET (v[i].fd, &set_write )) HE_BIT_SET (v[i].rcond, HE_IO_WRITE);
	if (FD_ISSET (v[i].fd, &set_except)) HE_BIT_SET (v[i].rcond, HE_IO_EXCEPT);
    }
    
    return res >= 0 ? res : errno == EINTR ? -1 : -2;
}


/*
 * Appel par HeIOLoop si erreur EBADF (mauvais fd).
 * Cherche le mauvais fd dans la liste et le supprime.
*/

void HeIOCorrectEBADF ()
{
    int i, res, fd;
    fd_set set_read;
    struct timeval tp;
    
    /* Pour retour immdiat de select() */
    tp.tv_sec = tp.tv_usec = 0;
    
    for (i = 0; i < he_io_list.nb; i++) {
    
        fd = ((He_io_struct*)he_io_list.v)[i].fd;
    
	FD_ZERO (&set_read);
        FD_SET (fd, &set_read);
        
	res = select (fd+1, &set_read, NULL, NULL, &tp);
        
        if (res < 0 && errno == EBADF) {  /* trouv */
        
            /* ATTENTION, l'appel  HeRemoveIO() va modifier he_io_list ;
             * c'est pourquoi on quitte immdiatement la fonction. 
             * S'il y a encore un autre fd dfectueux, un nouveau 
             * cycle de HeIOLoop rappellera HeIOCorrectEBADF().
            */
            if (HeRemoveIO (fd) == 0)
                 HeError ("HeIOCorrectEBADF: fd %d removed\n", fd);
            else HeError ("HeIOCorrectEBADF: can't remove fd %d\n", fd);
            return;
        }
    }
}


/*
 * Appel des callback sur les fd marqus par HeIOPoll().
 *
 * On travaille sur une copie de la liste des fd pour viter tout effet de
 * bord sur he_io_list pendant l'appel des callbacks.
*/

void HeIOProcess ()
{
    int i, j, nfd = 0;
    int *vfd = NULL;
    He_io_struct *elt;
   
    /* Aucun fd n'est marqu */
    if (he_io_list.nb == 0) return;
    
    /* Cration de la liste des fd */
    vfd = malloc (he_io_list.nb * sizeof(int));
    if (vfd == NULL) {
	HeError ("HeIOProcess: malloc error\n");
        return;
    }
    
    /* Copie de la liste des fd */
    for (i = 0; i < he_io_list.nb; i++) {
	elt = &((He_io_struct*)he_io_list.v)[i];
        if (elt->rcond != HE_IO_NULL)
            vfd[nfd++] = elt->fd;
    }
    
    /* Appel des callbacks */
    for (j = 0; j < nfd; j++) {
    
        /* On vrifie que vfd[j] est encore dans he_io_list 
         * (il a pu tre supprim entre-temps)
        */
        i = HeIOFind (vfd[j]);
        if (i < 0) continue;
	elt = &((He_io_struct*)he_io_list.v)[i];
        
        /* On vrifie que elt->fd est marqu (il a pu tre supprim puis
         * rinsr entre-temps,auquel cas rcond serait rinitialis)
        */
	if (elt->rcond != HE_IO_NULL)
            (elt->proc)(elt->fd, elt->rcond, elt->data);
    }
    
    /* Destruction liste des fd */
    free(vfd);
}


/*
 * Stoppe la boucle de HeIOLoop().
*/

void HeIOLoopStop ()
{
    he_ioloop_flag = FALSE;
}


/*
 * Boucle principale pseudo-infinie de scrutation de Helium.
 *
 * RQ : - le display X doit imprativement tre ouvert pendant toute 
 *        la dure de l'appel.
 *      - pour stopper la boucle on appelle HeIOLoopStop() depuis HeQuit.
*/

void HeIOLoop ()
{
    int res, timeout;
    struct timeval tv;

    while (he_ioloop_flag) {
    
        /* Timeout */
	timeout = HeGetNextTimeout (&tv);
    
        /* Attente bloquante */
        res = HeIOPoll (timeout == -1 ? NULL : &tv);
        
        /* Scrutation fd, X11, timeout et signaux.
         *
         *   - HeIOProcess() est appel si il y a des fd marqus par HeIOPoll().
         *     X11 est gr par ce mcanisme  partir de son fd.
         *
         *   - HeTimerProcess() est appel si HeIOPoll() est sorti  cause du
         *     timeout ; il est aussi appel aprs HeIOProcess(), pour tester si
         *     un timeout est arriv  chance pendant son excution.
         *
         *   - HeSignalProcess() est appel si HeIOPoll() est sorti  cause d'un
         *     signal Unix ; il est aussi appel si HeIOProcess() ou 
         *     HeTimerProcess() ont t appels, pour tester si un signal est
         *     arriv pendant leur excution. 
         *     RQ : sur certains systmes, le select() de HeIOPoll() n'est pas
         *          interrompu par un signal ; c'est pourquoi HeSignalNotify
         *          envoie un ClientMessage vide pour interrompre select(),
         *          qui sort avec res > 0 : c'est une raison supplmentaire
         *          pour appeler HeSignalProcess() quand res >= 1.
        */
        if (res >   0) HeIOProcess ();
        if (res >=  0) HeTimerProcess ();
        if (res >= -1) HeSignalProcess();
        
        if (res <  -1) {
            int e = errno;  /* HeError pourrait le modifier */
            HeError ("HeIOLoop: %s\n", strerror(e)); 
            /* Supprime mauvais fd, sinon select() bouclerait sur l'erreur */          
            if (e == EBADF) HeIOCorrectEBADF();
            continue;
        }
    
        /* Les update sont diffrs pour optimiser l'affichage. Les destroy
         * sont diffrs pour que les callbacks puissent etre appelees en
         * toute securite sans tester si les noeuds existent encore aprs.
         */
        HeUpdateFlush ();
        HeDestroyFlush ();
        
        /* Vide le tampon des requtes X11 et ralise l'affichage */
        XFlush (he_display);
    }
}



