/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/selection.c - 27/07/1999
 *
 * Selections XA_PRIMARY et CLIPBOARD
*/

#include <helium.h>


static int he_sel_owner = 0, he_sel_requestor = 0,
           he_own_primary = FALSE, he_own_clipboard = FALSE, 
           he_clipboard_len = 0;
           
static char *he_clipboard_str = NULL;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Renvoie le noeud qui a la selection.
*/

He_node *HeGetSelection ()
{
    return HeGetNodeFromId (he_sel_owner);
}


/*
 * Annule la selection actuelle.
*/

void HeClearSelection ()
{
    if (he_sel_owner != 0) {
        He_node *old = HeGetSelection();
        if (old != NULL) {
            HE_BIT_CLR (old->status, HE_STA_SELECTION);
            HeUpdateMark (old, HE_UPD_DRAW);
        }
        he_sel_owner = 0;
    }
}


/*
 * Annule la selection actuelle,
 * puis donne la selection a hn si non NULL. 
*/

void HeSetSelection (He_node *hn)
{
    /* Test si deja selectionne */
    if (hn != NULL && HE_BIT_ISSET (hn->status, HE_STA_SELECTION)) return;
    
    /* Perte de l'ancienne selection */
    HeClearSelection ();
    
    /* Nouvelle selection */
    if (hn == NULL) return;
    he_sel_owner = hn->id;
    HE_BIT_SET (hn->status, HE_STA_SELECTION);
    HeUpdateMark (hn, HE_UPD_DRAW);
    
    /* L'appli devient proprietaire de l'atome PRIMARY */
    HeSetPrimaryOwner ();
}


/*
 * Le noeud hn demande la selection a une autre appli.
*/

void HeImportSelection (He_node *hn, Atom selection, Atom target, Atom prop)
{
    if (hn == NULL) { he_sel_requestor = 0; return; }
    
    he_sel_requestor = hn->id;
    
    XConvertSelection (he_display, selection, target, prop, 
	he_main_frame->win, CurrentTime);
}


/*
 * Libere le string memorisant le clipboard
*/

void HeClearClipboard ()
{
    free (he_clipboard_str); 
    he_clipboard_str = NULL; 
    he_clipboard_len = 0;
}


/*
 * Memorise un string pour le clipboard
*/

void HeSetClipboardString (const char *str, int len)
{
    HeClearClipboard ();

    if (str == NULL || len <= 0) return;
    
    he_clipboard_str = malloc (len+1);
    
    if (he_clipboard_str == NULL) return;
    
    he_clipboard_len = len;
    memcpy (he_clipboard_str, str, len);
    he_clipboard_str[len] = 0;
    
    /* L'appli devient proprietaire de l'atome CLIPBOARD */
    HeSetClipboardOwner ();
}


/*
 * Donne le string memorise dans le clipboard
*/

void HeGetClipboardString (char **str, int *len, int *internal)
{
    *str = he_clipboard_str;
    *len = he_clipboard_len;
    *internal = he_own_clipboard;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Demande au serveur X que l'application deviennent proprietaire de
 * l'atome PRIMARY.
 *
 * Appele par HeSetSelection().
*/

void HeSetPrimaryOwner ()
{
    /* Si on est deja proprietaire, on sort sans MAJ time */
    if (he_own_primary) return;
    
    /* Demande a etre le proprietaire de l'atome PRIMARY */
    XSetSelectionOwner(he_display, XA_PRIMARY, he_main_frame->win, CurrentTime);
    
    /* Verifie si la demande est acceptee */
    if (XGetSelectionOwner(he_display, XA_PRIMARY) == he_main_frame->win)
	he_own_primary = TRUE;
}


/*
 * Demande au serveur X que l'application deviennent proprietaire de
 * l'atome CLIPBOARD, ou met ajour le time.
 *
 * Appele par HeSetClipboardString().
*/

void HeSetClipboardOwner ()
{
    /* Demande a etre le proprietaire de l'atome CLIPBOARD, MAJ time */
    XSetSelectionOwner(he_display, he_atom_clipboard, he_main_frame->win, 
	CurrentTime);
    
    /* Verifie si la demande est acceptee */
    he_own_clipboard = 
      (XGetSelectionOwner(he_display, he_atom_clipboard) == he_main_frame->win)
	? TRUE : FALSE;
}


/*
 * Une autre appli X demande le clipboard.
 * Renvoie TRUE si on a pu repondre.
 *
 * Appele par HeXSelRequest().
*/

int HeExportClipboard (Atom sel_type, Atom target, Atom *type, 
	Uchar **value, int *length, int *format, int *free_value)
{
    
    if (sel_type != he_atom_clipboard) return FALSE;
    
    /* 
     * On nous demande le texte selectionne'
    */
    if (target == XA_STRING) {
    
        *type = XA_STRING;
        if ( he_clipboard_str != NULL && he_clipboard_len > 0)
        { 
            *value = (Uchar *) he_clipboard_str; 
            *length = he_clipboard_len; 
        }
        else { 
            *value = (Uchar *) he_empty_str; 
            *length = 0;
        }
        *format = 8;
        *free_value = FALSE;
        return TRUE;
    }
    
    /*
     * On nous demande la liste des targets qu'on peut traiter
    */
    if (target == he_atom_targets) {
    
    /*  RQ : ne plait pas a cc de SGI :
        Atom liste[] = { he_atom_targets, XA_STRING };
        int taille = sizeof(liste) / sizeof(Atom);
    */  
        Atom liste[2]; int taille = 2;
        liste[0] = he_atom_targets; liste[1] = XA_STRING;
        
        *type = XA_ATOM;
        *value = (Uchar *) liste;
        *length = taille;
        *format = 32;
        *free_value = FALSE;
        return TRUE;
    }
    
    return FALSE;
}


/*
 * Perte d'une selection : une autre appli est devenue proprietaire.
 *
 * Appele par HeEventDispatch().
*/

void HeXSelClear (XSelectionClearEvent *ev)
{
#ifdef DEBUG
    printf ("HeXSelClear :\n  selection = ");
    HePrintAtomName (ev->selection); printf ("\n");
#endif

    if (ev->selection == XA_PRIMARY) {
	he_own_primary = FALSE;
	HeClearSelection ();
    } else if (ev->selection == he_atom_clipboard) {
	he_own_clipboard = FALSE;
	HeClearClipboard ();
    }
}


/*
 * Une autre appli demande la selection.
 *
 * Appele par HeEventDispatch().
*/

void HeXSelRequest (XSelectionRequestEvent *ev)
{
    XEvent send;
    Atom type = 0;
    Uchar *value = NULL;
    int succes = FALSE, length = 0, format = 8, free_value = FALSE;
    
#ifdef DEBUG
    printf ("HeXSelRequest :\n");
    HePrintSelAtoms (ev->selection, ev->target, ev->property); 
#endif
    
    if (ev->selection == he_atom_clipboard) {
    
        succes = HeExportClipboard (ev->selection, ev->target,
	    &type, &value, &length, &format, &free_value);
    
    } else {
    
	He_node *hn = HeGetSelection ();
    
	/* Appel methode */
	if (hn != NULL)
            succes = (hn->meth->sel_export)(hn, ev->selection, ev->target,
		&type, &value, &length, &format, &free_value);
    }
        
    /* Ecrit la propriete chez le requestor */
    if (succes) {
        XChangeProperty (he_display, ev->requestor, 
            ev->property, type, format, PropModeReplace,
            value, length);    
        if (free_value) free(value);
    }
    
    /* Envoie l'evenement reponse au requestor */
    send.xselection.type      = SelectionNotify;
    send.xselection.display   = he_display;
    send.xselection.requestor = ev->requestor;
    send.xselection.selection = ev->selection;
    send.xselection.target    = ev->target;
    send.xselection.property  = succes ? ev->property : None;
    send.xselection.time      = ev->time;
    XSendEvent (he_display, ev->requestor, False, 0L, &send);
}


/*
 * On recoit la selection demandee a une autre appli.
 *
 * Appele par HeEventDispatch().
*/

void HeXSelNotify (XSelectionEvent *ev)
{
    He_node *hn;
    Atom type = 0;
    Ulong nitems = 0;
    Uchar *data = NULL;
    int format = 0, err, free_data = TRUE;

#ifdef DEBUG
    printf ("HeXSelNotify :\n");
    HePrintSelAtoms (ev->selection, ev->target, ev->property); 
#endif
    
    /* On recupere la propriete et on l'enleve du Window (delete=True)*/
    err = HeAtomGetProp (ev->requestor, ev->property, True,
	&type, &format, &nitems, &data);
        
#ifdef DEBUG
    /* Affiche la liste des targets acceptes */
    if (err == 0 && ev->property != None && 
        ev->target == he_atom_targets && type == XA_ATOM && format == 32) 
            HePrintSelTargets ((Atom*) data, nitems);
#endif
 
   /* On recupere le demandeur de la selection */
    hn = HeGetNodeFromId (he_sel_requestor);
        
    /* Appel methode */
    if (hn != NULL) 
	(hn->meth->sel_import)(hn, ev->selection, ev->target,
            (err == 0) ? ev->property : None, 
            type, format, nitems, data, &free_data);

    /* Libere la memoire */
    if (free_data && data) XFree(data);
}


/*
 * Pour le debugage.
*/

void HePrintSelAtoms (Atom selection, Atom target, Atom property)
{
    printf ("  selection = "); HePrintAtomName (selection);
    printf ("  target = ");    HePrintAtomName (target);
    printf ("  property = ");  HePrintAtomName (property);
    printf ("\n");
}

void HePrintSelTargets (Atom liste[], int nb)
{
    int i;
    printf ("( ");
    for (i = 0; i < nb; i++) { HePrintAtomName (liste[i]); printf (" "); }
    printf (")\n");
}

