/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/signal.c - 21/12/1999
 *
 * Gestion des signaux Unix
*/

#include <helium.h>


/*
 * En principe, NSIG est dfini dans /usr/include/sys/signal.h
*/
#ifndef NSIG
#define NSIG 32
#endif
static He_sig_proc he_signal_proc[NSIG];
static void *he_signal_data[NSIG];
static int he_signal_hop[NSIG], he_signal_nhop;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Connecte une callback sur un signal Unix sig.
 * Un nouvel appel ecrase le precedent. 
 * Renvoie 0 succes, -1 echec.
 *
 * La callback sera appele d'un endroit "sr" dans HeMainLoop, d'o on pourra
 * faire ce qu'on veut. La callback recevra sig et data.
 * Prototype :  void proc (int sig, void *data)
 *
 * Helium ne se rserve aucun signal ; en particulier, le mcanisme des timeout
 * est gr par un select() et non par un SIGALRM.
*/

int HeAddSignal (int sig, He_sig_proc proc, void *data)
{
    if (sig < 1 || sig >= NSIG)
    {
	HeError ("HeAddSignal(%d,,) failed: signal interval is [1..%d[\n", 
            sig, NSIG);
	return -1;
    }
    if (proc == NULL)
    {
	HeError ("HeAddSignal(%d,proc,) failed: proc is NULL\n", sig);
	return -1;
    }
    
    he_signal_proc[sig] = proc;
    he_signal_data[sig] = data;
    
    if (HeSignalInstall (sig, HeSignalNotify) < 0)
    {
	HeError ("HeAddSignal(%d,,) failed: can't handle signal: %s\n", 
            sig, strerror(errno));
	he_signal_proc[sig] = NULL;
	he_signal_data[sig] = NULL;      
        return -1;
    }
    
    return 0;
}


/*
 * Supprime une callback sur un signal Unix sig.
 * Le handler systme par dfaut est rtabli. 
 * Renvoie 0 succes, -1 echec.
*/

int HeRemoveSignal (int sig)
{
    if (sig < 1 || sig >= NSIG)
    {
	HeError ("HeRemoveSignal(%d) failed: signal interval is [1..%d[\n", 
            sig, NSIG);
	return -1;
    }
    if (he_signal_proc[sig] == NULL)
    {
	HeError ("HeRemoveSignal(%d) failed: signal was never set\n", sig);
	return -1;
    }
    
    he_signal_proc[sig] = NULL;
    he_signal_data[sig] = NULL;
    
    /* RQ : SIG_DFL est le handler par defaut du systeme */
    if (HeSignalInstall (sig, SIG_DFL) < 0)
    {
	HeError ("HeRemoveSignal(%d) failed: can't handle signal: %s\n", 
            sig, strerror(errno));
        return -1;
    }
    
    return 0;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Init le module.
*/

void HeSignalInit ()
{
    int i;
    
    for (i = 0; i < NSIG; i++) {
	he_signal_proc[i] = NULL;
        he_signal_data[i] = NULL;
        he_signal_hop[i] = 0;
    }
    he_signal_nhop = 0;
}


/*
 * Mise en place d'un handler de signal.
 *
 * Retour 0 succes, -1 erreur.
 *
 * SYSTEME : signal() n'agit pas pareil sous BSD, sysV et POSIX !!
 * La solution est d'utiliser sigaction() en le parametrant tel que :
 *
 *	- le signal recu est masque'
 *	- le handler est conserve'
 *	- les appels systeme bloquants sont repris (sauf select(), cf RQ1).
 *
 * RQ1 : l'appel select() est interrompu sous Linux et Solaris (c'est ce
 *       qu'on veut), mais pas sur SGI !! 
 *       La solution consiste  envoyer un ClientMessage lors du 1er signal, 
 *       pour sortir du select() par fdX.
*/

int HeSignalInstall (int sig, void (*sig_proc)(int) )
{
    struct sigaction action;

#   ifndef SA_RESTART
#   define SA_RESTART 0x0000	/* Pour Sun 4.1.3 */
#   endif
    sigemptyset (&action.sa_mask);
    action.sa_handler = sig_proc;
    action.sa_flags = SA_RESTART;

    if (sigaction (sig, &action, NULL) < 0)
	return -1;

    return 0;
}


/*
 * Handler de signal commun  tous les signaux. Le but est d'enregistrer le
 * signal pour traitement ultrieur, et de faire sortir du select().
*/

void HeSignalNotify (int sig)
{
    if (sig < 1 || sig >= NSIG) return;
    if (he_signal_proc[sig] == NULL) return;
    
    /* Les signaux envoyes avant ou apres HeMainLoop sont 
       silencieusement ignores. */
    if (!he_main_frame) return;
    
    /* Si c'est le premier signal enregistr, on envoie un ClientMessage
       pour sortir du select(), cf RQ1 dans HeSignalInstall() */
    if (he_signal_nhop == 0) {
        long l[5];
        HeSendLongMessage (l, he_atom_sig);
    }
    
    /* On marque le signal pour HeSignalProcess() */
    he_signal_hop[sig]++; he_signal_nhop++;
}


/*
 * Appel des callback pour tous les signaux capts. Si un signal a t capt
 * plusieurs fois (hop > 1) la callback n'est appele qu'une seule fois.
*/

void HeSignalProcess ()
{
    int i;
    
    if (he_signal_nhop == 0) return;
    
    for (i = 0; i < NSIG; i++)
        if (he_signal_hop[i] > 0 && he_signal_proc[i] != NULL)
        {
            (he_signal_proc[i])(i, he_signal_data[i]);
            he_signal_hop[i] = 0;  /* RAZ */
        }
    
    he_signal_nhop = 0;  /* RAZ */
}


 
