/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/tip.c - 30/01/2002
 *
 * Affichage des conseils dans des bubble.
*/

#include <helium.h>


static int  he_tip_target_id = 0,	/* widget dont on va afficher l'aide */
            he_tip_bubble_id = 0,	/* bubble qui sert  afficher l'aide */
            he_tip_timer_han = 0,	/* handle du timer */
            he_tip_x = 0,		/* Coordonnes de la bulle */
            he_tip_y = 0,
            he_tip_leave_id = 0;	/* id du dernier Leave */
static Time he_tip_leave_time = 0;	/* time du dernier Leave */


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/


/*
 * Attache un message d'aide au widget hn.
 * Le message peut contenir des '\n'.
 *
 * Si le widget n'est pas capable de mmoriser le tip, 
 * cette fonction choue silencieusement.
 *
 * Si le message est actuellement affich dans une bulle,
 * la bulle est mise  jour si label != NULL, efface sinon.
*/

void HeSetTip (He_node *hn, char *label)
{
    He_tip_info *tip_info;
    
    if (hn == NULL) return; 
    
    /* Test si la mmorisation est possible */
    tip_info = (hn->meth->tip)(hn);
    if (tip_info == NULL) return;
    
    /* Mmorisation de label */
    free (tip_info->label);
    tip_info->label = (label == NULL) ? NULL : strdup(label);
    
    /* Mise  jour de la bulle si elle est en cours d'affichage */
    if (he_tip_target_id == hn->id && he_tip_timer_han == 0) {
	He_node *bubble = HeGetNodeFromId (he_tip_bubble_id);
	if (HeGetShow (bubble)) {
	    if (label == NULL)
                HeSetShow (bubble, FALSE);
            else {
                HeSetBubbleLabel (bubble, label);
	        HeShowBubbleXYF (bubble, hn, he_tip_x, he_tip_y,
		    !HE_BIT_ISSET (hn->meth->props, HE_PM_WIN_OWNER));
            }
        }
    }
}


/*
 * Renvoie le message d'aide attach au widget hn.
*/

char *HeGetTip (He_node *hn)
{
    He_tip_info *tip_info;
    
    if (hn == NULL) return he_empty_str; 
    
    tip_info = (hn->meth->tip)(hn);
    if (tip_info == NULL) return he_empty_str;
    
    return tip_info->label ? tip_info->label : he_empty_str;
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialise les champs de tip_info.
*/

void HeTipInitInfo (He_tip_info *tip_info)
{
    if (tip_info == NULL) return;
    tip_info->label = NULL;
}


/*
 * Libre les champs de tip_info.
*/

void HeTipDestroyInfo (He_tip_info *tip_info)
{
    if (tip_info == NULL) return;
    free (tip_info->label);
}


/*
 * Gestion des evenements : armement/dsarmement des bulles d'aide. 
 * On suppose hn et hev non nuls.
 * Renvoie 0 si succs, -1 si erreur.
 *
 * Appele' par HeEventDispatch()
*/

int HeTipEvent (He_node *hn, He_event *hev)
{
    switch (hev->type) {

	case EnterNotify :
            if (hev->sb > 0) break;
            if ( he_tip_leave_id == hn->id && 
                 hev->time - he_tip_leave_time < 200 ) break;
            HeTipStop ();
	    HeTipStart (hn); 
            break;
        
        case LeaveNotify :
            he_tip_leave_time = hev->time;
            he_tip_leave_id = hn->id;
            HeTipStop ();
            break;
    
        case KeyPress :
        case KeyRelease :
        case ButtonPress :
        case ButtonRelease :
            HeTipStop ();
            break;
    
    }
    return 0;
}


/*
 * Supprime le timer et efface la bulle d'aide courante.
*/

void HeTipStop ()
{
    /* Suppression du timer */
    if (he_tip_timer_han != 0) {
        HeRemoveTimeout (he_tip_timer_han);
        he_tip_timer_han = 0;
    }
    
    /* Effacement bulle d'aide */
    if (he_tip_bubble_id != 0) {
	He_node *bubble = HeGetNodeFromId (he_tip_bubble_id);
	if (bubble != NULL) HeSetShow (bubble, FALSE);
    }
}


/*
 * Rarme le timer si hn != NULL et s'il y a un tip.
*/

void HeTipStart (He_node *hn)
{
    He_tip_info *tip_info;
    
    if (hn == NULL || !HE_BIT_ISSET (hn->status, HE_STA_ACTIVE)) return;
    
    tip_info = (hn->meth->tip)(hn);
    if (tip_info == NULL || tip_info->label == NULL) return;

    he_tip_timer_han = HeAddTimeout (500, HeTipTimerProc, NULL);
    he_tip_target_id = hn->id;
}


/*
 *  Callback pour le timer, qui affiche la bulle d'aide.
*/

int HeTipTimerProc (int h, void *data)
{
    He_node *target, *bubble;
    
    /* RAZ du handle du timer dans tous les cas */
    he_tip_timer_han = 0;
    
    /* Retrouve la cible */
    target = HeGetNodeFromId (he_tip_target_id);
    if (target == NULL) return FALSE;
    
    /* Retrouve ou cre la bulle d'aide */
    bubble = HeGetNodeFromId (he_tip_bubble_id);
    if (bubble == NULL) {
        he_tip_bubble_id = 0;
        bubble = HeCreateBubble ();
        if (bubble == NULL) return FALSE;
        he_tip_bubble_id = bubble->id;
    }
    
    /* Mmorise les coordonnes de la souris relatives  target */
    if (HeQueryPointerCoords (target, &he_tip_x, &he_tip_y) == FALSE) 
        return FALSE;
        
    /* Affiche la bulle d'aide pour la cible */
    HeSetBubbleLabel (bubble, HeGetTip(target));
    HeShowBubbleXYF (bubble, target, he_tip_x, he_tip_y,
	!HE_BIT_ISSET (target->meth->props, HE_PM_WIN_OWNER));
    
    /* Suppression du timer */
    return FALSE ;
}

