/*
 * The Helium Toolkit
 *
 * Copyright (C) 1996-2000 Edouard Thiel <Edouard.Thiel@lif.univ-mrs.fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License; see http://www.lif-sud.univ-mrs.fr/~thiel/helium/LICENSE
*/

/*
 * gui/update.c - 28/03/2000
 *
 * Affichages differes
*/

#include <helium.h>


static int he_updating_flag = FALSE; 

static He_vlist he_upd_list;


/*------------------- I N T E R F A C E - P U B L I Q U E --------------------*/

/*
 * Realisation immediate de tous les affichages en attente.
 * Dans la philosophie de l'interface, a utiliser le moins possible.
*/

void HeFlushDisplay ()
{
    HeUpdateFlush (); 
    XFlush (he_display);
}


/*
 * Mise  jour immdiate de la gomtrie et de l'affichage de hn
 * (sans faire de XFlush).
*/

void HeUpdate (He_node *hn)
{
    He_upd_info *ui;
    
    if (hn == NULL || hn->upd_index < 0 || hn->upd_index >= he_upd_list.nb)
        return;
    
    /* Interdit d'appeler HeUpdateInfo() ou HeUpdateMark() */
    he_updating_flag = TRUE;
    
    /* On retrouve le ui de hn */
    ui = &((He_upd_info*)he_upd_list.v)[hn->upd_index];
    
    /* Appel mthode update */
    (hn->meth->update)(hn, ui);
    
    /* RAZ */
    hn->upd_index = -1;
    ui->hn = NULL;
    
    he_updating_flag = FALSE;
}


/*
 * Provoque un raffichage diffr.
*/

void HePostRepaint (He_node *hn)
{
    /* On aurait pu appeler directement HeSendExpose() ; mais l'evenement
     * Expose serait stoqu directement dans la Queue X, sans passer par le fd
     * de X. Cela ferait boucler indfiniment HeEventInputProc(), ce qui
     * bloquerait la boucle de HeIOLoop() !!
     * La solution est simple : on met un flag HE_UPD_EXPOSE qui sera trait 
     * pendant HeUpdateFlush(), c'est--dire APRES HeEventInputProc(). 
    */
    HeUpdateMark (hn, HE_UPD_EXPOSE);
}


/*-------------------- F O N C T I O N S - P R I V E E S ---------------------*/

/*
 * Initialisation generale du module.
*/

void HeUpdateInit ()
{
   /* Liste de He_upd_info non trie
    * stockage dans l'ordre d'apparition (index fixe pour un elt).
   */
    HeVListInit (&he_upd_list, FALSE, 128, sizeof(He_upd_info), NULL);
}


/*
 * Init des champs de ui, sauvegarde des coords.
*/

void HeUpdateInitElt (He_node *hn, He_upd_info *ui)
{
    ui->hn = hn;
    if (hn == NULL) return;
    ui->xb = hn->xb;
    ui->yb = hn->yb;
    ui->xm = hn->xm;
    ui->ym = hn->ym;
    ui->mask = 0L;
}


/*
 * Marque hn dans he_upd_list, 
 * et ajoute une valeur au masque avec un OR.
 *
 * Toujours appeler AVANT de changer une coord
*/

void HeUpdateMark (He_node *hn, Ulong val)
{
    He_upd_info *ui = HeUpdateInfo (hn);
    if (ui == NULL) return; 
    HE_BIT_SET (ui->mask, val);
}


/*
 * Marque hn dans he_upd_list.
 * Renvoie ui*
*/

He_upd_info *HeUpdateInfo (He_node *hn)
{
    if (he_updating_flag) {
	HeError ("HeUpdateInfo: called in an update_meth()\n");
	return NULL;
    } 
    
    if (hn == NULL) return NULL;
    
    /* Si ce test est vrai alors hn n'est pas dans la liste
     * sinon hn est dj dans la liste.
    */
    if (hn->upd_index < 0 || hn->upd_index >= he_upd_list.nb)
    {
        He_upd_info ui;
        int index;
        
        /* Init des champs de ui */
	HeUpdateInitElt(hn, &ui);
    
        /* Insertion de ui par recopie */
        index = HeVListInsert (&he_upd_list, &ui);
        
        /* Test russite */
        if (index < 0) return NULL;
        
        /* Mmorisation index pour accs direct */
        hn->upd_index = index;
    }

    /* Dans tous les cas */
    return &((He_upd_info*)he_upd_list.v)[hn->upd_index];
}


/*
 * Ralisation des Update
 *
*/

void HeUpdateFlush ()
{
    int i;
    He_node *hn;
    
    if (he_upd_list.nb == 0) return;
    
#   ifdef DEBUG
    printf ("HeUpdateFlush : nb = %d  max = %d\n", 
            he_upd_list.nb, he_upd_list.max);
#   endif

    /* Interdit d'appeler HeUpdateInfo() ou HeUpdateMark() */
    he_updating_flag = TRUE;
    
    for (i = 0; i < he_upd_list.nb; i++)
    {
        He_upd_info *ui = &((He_upd_info*)he_upd_list.v)[i];
        
	/* On rcupere le hn de ui */
	hn = ui->hn;
        if (hn == NULL) continue;
        
        /* Appel mthode update */
        (hn->meth->update)(hn, ui);
        
        /* RAZ hn */
        hn->upd_index = -1;  
    }
    
    /* RAZ liste */
    HeVListRAZ (&he_upd_list);
    
    he_updating_flag = FALSE;
}


/*
 * M-a-j geometrie d'un widget base' sur un sub-window.
 *
 * Appele' par mthodes Update 
*/

void HeUpdateWinGeom (He_node *hn, He_upd_info *ui)
{
    unsigned int   mc = 0;  /* mask changes */
    XWindowChanges wc;

    if (hn == NULL || ui == NULL || hn->win == 0) return; 

    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMX)) {
        HE_BIT_SET (mc, CWX); wc.x = hn->xb;
    }
    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMY)) {
        HE_BIT_SET (mc, CWY); wc.y = hn->yb;
    }
    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMW)) {
        HE_BIT_SET (mc, CWWidth);  wc.width  = hn->xm;
    }
    if (HE_BIT_ISSET (ui->mask, HE_UPD_DIMH)) {
        HE_BIT_SET (mc, CWHeight); wc.height = hn->ym;
    }

    if (mc != 0) 
	/* C'est un sub-window */
	XConfigureWindow (he_display, hn->win, mc, &wc);
}


/*
 * M-a-j geometrie d'un gadget : on efface avec les anciennes coords, 
 * qui sont stockes dans ui.
 * Appele' par mthodes Update.
*/

void HeUpdateDrawOldBg (He_node *hn, He_upd_info *ui)
{
    if (hn == NULL || ui == NULL) return; 

    if (HE_BIT_ISSET (ui->mask, HE_UPD_ANYDIM) && ui->xm > 0 && ui->ym > 0) {
	XSetForeground (he_display, he_gc_draw, he_col_bg);
        XFillRectangle (he_display, hn->win, he_gc_draw,
            ui->xb, ui->yb, ui->xm, ui->ym);
    }
}


